@@ -3,7 +3,7 @@
 ##
 ##    Functions for generating random point patterns
 ##
-##    $Revision: 4.105 $   $Date: 2022/01/04 05:30:06 $
+##    $Revision: 4.113 $   $Date: 2023/01/06 11:57:39 $
 ##
 ##    runifpoint()      n i.i.d. uniform random points ("binomial process")
 ##    runifdisc()       special case of disc (faster)
@@ -43,13 +43,13 @@ runifdisc <- function(n, radius=1, centre=c(0,0), ..., nsim=1, drop=TRUE)
   stopifnot(radius > 0)
   if(!missing(nsim)) {
     check.1.integer(nsim)
-    stopifnot(nsim >= 1)
+    stopifnot(nsim >= 0)
   }
   disque <- disc(centre=centre, radius=radius, ...)
   twopi <- 2 * pi
   rad2 <- radius^2
   result <- vector(mode="list", length=nsim)
-  for(isim in 1:nsim) {
+  for(isim in seq_len(nsim)) {
     theta <- runif(n, min=0, max=twopi)
     s <- sqrt(runif(n, min=0, max=rad2))
     result[[isim]] <- ppp(centre[1] + s * cos(theta),
@@ -67,7 +67,7 @@ runifpoint <- function(n, win=owin(c(0,1),c(0,1)),
 {
   if(!missing(nsim)) {
     check.1.integer(nsim)
-    stopifnot(nsim >= 1)
+    stopifnot(nsim >= 0)
   }
   
   if(missing(n) && missing(win) && !is.null(ex)) {
@@ -109,7 +109,7 @@ runifpoint <- function(n, win=owin(c(0,1),c(0,1)),
            ypix <- rxy$y
            ## make a list of nsim point patterns
            result <- vector(mode="list", length=nsim)
-           for(isim in 1:nsim) {
+           for(isim in seq_len(nsim)) {
              ## select pixels with equal probability
              id <- sample(seq_along(xpix), n, replace=TRUE)
              ## extract pixel centres and randomise within pixels
@@ -121,7 +121,7 @@ runifpoint <- function(n, win=owin(c(0,1),c(0,1)),
          polygonal={
            ## make a list of nsim point patterns
            result <- vector(mode="list", length=nsim)
-           for(isim in 1:nsim) {
+           for(isim in seq_len(nsim)) {
              ## rejection method
              ## initialise empty pattern
              x <- numeric(0)
@@ -172,7 +172,7 @@ runifpoispp <- function(lambda, win = owin(c(0,1),c(0,1)), ...,
     stop("Intensity lambda must be a single finite number >= 0")
   if(!missing(nsim)) {
     check.1.integer(nsim)
-    stopifnot(nsim >= 1)
+    stopifnot(nsim >= 0)
   }
 
   if(lambda == 0) {
@@ -188,7 +188,7 @@ runifpoispp <- function(lambda, win = owin(c(0,1),c(0,1)), ...,
   meanN <- lambda * area(box)
   
   result <- vector(mode="list", length=nsim)
-  for(isim in 1:nsim) {
+  for(isim in seq_len(nsim)) {
     n <- rpois(1, meanN)
     if(!is.finite(n))
       stop(paste("Unable to generate Poisson process with a mean of",
@@ -204,12 +204,12 @@ runifpoispp <- function(lambda, win = owin(c(0,1),c(0,1)), ...,
 }
 
 rpoint <- function(n, f, fmax=NULL,
-                   win=unit.square(), ..., giveup=1000,verbose=FALSE,
+                   win=unit.square(), ..., giveup=1000, warn=TRUE, verbose=FALSE,
                    nsim=1, drop=TRUE, forcewin=FALSE) {
   
   if(missing(f) || (is.numeric(f) && length(f) == 1))
     ## uniform distribution
-    return(runifpoint(n, win, giveup, nsim=nsim, drop=drop))
+    return(runifpoint(n, win, giveup=giveup, warn=warn, nsim=nsim, drop=drop))
   
   ## non-uniform distribution....
   if(!is.function(f) && !is.im(f))
@@ -219,7 +219,7 @@ rpoint <- function(n, f, fmax=NULL,
 
   if(!missing(nsim)) {
     check.1.integer(nsim)
-    stopifnot(nsim >= 1)
+    stopifnot(nsim >= 0)
   }
   
   if(is.im(f)) {
@@ -258,7 +258,7 @@ rpoint <- function(n, f, fmax=NULL,
     ppix <- as.vector(f$v[M]) ## not normalised - OK
     ## generate
     result <- vector(mode="list", length=nsim)
-    for(isim in 1:nsim) {
+    for(isim in seq_len(nsim)) {
       ## select pixels
       id <- sample(npix, n, replace=TRUE, prob=ppix)
       ## extract pixel centres and randomise location within pixels
@@ -306,7 +306,7 @@ rpoint <- function(n, f, fmax=NULL,
   box <- boundingbox(win)
 
   result <- vector(mode="list", length=nsim)
-  for(isim in 1:nsim) {
+  for(isim in seq_len(nsim)) {
 
     ## initialise empty pattern
     X <- ppp(numeric(0), numeric(0), window=win)
@@ -369,7 +369,7 @@ rpoispp <- function(lambda, lmax=NULL, win = owin(), ...,
 
   if(!missing(nsim)) {
     check.1.integer(nsim)
-    stopifnot(nsim >= 1)
+    stopifnot(nsim >= 0)
   }
   
   if(missing(lambda) && is.null(lmax) && missing(win) && !is.null(ex)) {
@@ -415,7 +415,7 @@ rpoispp <- function(lambda, lmax=NULL, win = owin(), ...,
     #'      runifpoispp checks 'lmax'
     result <- runifpoispp(lmax, win, nsim=nsim, drop=FALSE)
     #'      result is a 'ppplist' with appropriate names
-    for(isim in 1:nsim) {
+    for(isim in seq_len(nsim)) {
       X <- result[[isim]]
       if(X$n > 0) {
         prob <- lambda(X$x, X$y, ...)/lmax
@@ -424,9 +424,7 @@ rpoispp <- function(lambda, lmax=NULL, win = owin(), ...,
         result[[isim]] <- X[retain]
       }
     }
-    if(nsim == 1 && drop)
-       result <- result[[1L]]
-    return(result)
+    return(simulationresult(result, nsim, drop))
   }
 
   if(is.im(lambda)) {
@@ -456,7 +454,7 @@ rpoispp <- function(lambda, lmax=NULL, win = owin(), ...,
     } else {
       ## old code: thinning
       result <- runifpoispp(lmax, win, nsim=nsim, drop=FALSE)
-      for(isim in 1:nsim) {
+      for(isim in seq_len(nsim)) {
         X <- result[[isim]]
         if(X$n > 0) {
           prob <- lambda[X]/lmax
@@ -465,9 +463,7 @@ rpoispp <- function(lambda, lmax=NULL, win = owin(), ...,
           result[[isim]] <- X[retain]
         }
       }
-      if(nsim == 1 && drop)
-         return(result[[1L]])
-      return(result)
+      return(simulationresult(result, nsim, drop))
     }
   }
   stop(paste(sQuote("lambda"), "must be a constant, a function or an image"))
@@ -497,7 +493,7 @@ rMaternInhibition <- function(type,
   stopifnot(type %in% c(1,2))
   if(!missing(nsim)) {
     check.1.integer(nsim)
-    stopifnot(nsim >= 1)
+    stopifnot(nsim >= 0)
   }
   ## Resolve window class
   if(!inherits(win, c("owin", "box3", "boxx"))) {
@@ -519,7 +515,7 @@ rMaternInhibition <- function(type,
     bigbox <- if(stationary) grow.boxx(win, r) else win
     result <- rpoisppx(kappa, domain = bigbox, nsim = nsim, drop=FALSE)
   }
-  for(isim in 1:nsim) {
+  for(isim in seq_len(nsim)) {
     Y <- result[[isim]]
     nY <- npoints(Y)
     if(type == 1) {
@@ -635,7 +631,7 @@ rSSI <- function(r, n=Inf, win = square(1),
 
   #' start simulation 		    
   pstate <- list()
-  for(isim in 1:nsim) {
+  for(isim in seq_len(nsim)) {
     if(nsim > 1) pstate <- progressreport(isim, nsim, state=pstate)
     ## Simple Sequential Inhibition process
     ## fixed number of points
@@ -684,7 +680,7 @@ rSSI <- function(r, n=Inf, win = square(1),
 
 rPoissonCluster <-
   function(kappa, expand, rcluster, win = owin(c(0,1),c(0,1)), ...,
-           lmax=NULL, nsim=1, drop=TRUE, saveparents=TRUE)
+           nsim=1, drop=TRUE, saveparents=TRUE, kappamax=NULL)
 {
   ## Generic Poisson cluster process
   ## Implementation for bounded cluster radius
@@ -695,35 +691,45 @@ rPoissonCluster <-
   ## "..." are arguments to be passed to 'rcluster()'
   ##
 
-  ## Catch old argument name rmax for expand, and allow rmax to be
-  ## passed to rcluster (and then be ignored)
-  if(missing(expand) && !is.null(rmax <- list(...)$rmax)){
-      expand <- rmax
-      f <- rcluster
-      rcluster <- function(..., rmax) f(...)
-  }
-  win <- as.owin(win)
-  
   if(!missing(nsim)) {
     check.1.integer(nsim)
-    stopifnot(nsim >= 1)
+    stopifnot(nsim >= 0)
+    if(nsim == 0) return(simulationresult(list()))
   }
 
+  ## Catch old argument name rmax for expand, and allow rmax to be
+  ## passed to rcluster (and then be ignored)
+  if(missing(expand) && !is.null(rmax <- list(...)$rmax)) {
+    warning("outdated usage in rPoissonCluster: 'rmax' should be 'expand'")
+    expand <- rmax
+  }
+
+  rPoissonClusterEngine(kappa=kappa, expand=expand, rcluster=rcluster,
+                        win=win, nsim=nsim, drop=drop, saveparents=saveparents,
+                        kappamax=kappamax, ...)
+}
+
+rPoissonClusterEngine <- function(kappa, expand=rmax, rcluster, win, ..., 
+                                  nsim=1, drop=TRUE, saveparents=TRUE,
+                                  kappamax=lmax, lmax=NULL, rmax=NULL) {
   ## Generate parents in dilated window
+  win <- as.owin(win)
   frame <- boundingbox(win)
   dilated <- owin(frame$xrange + c(-expand, expand),
                   frame$yrange + c(-expand, expand))
+
   if(is.im(kappa) && !is.subset.owin(dilated, as.owin(kappa)))
     stop(paste("The window in which the image",
                sQuote("kappa"),
                "is defined\n",
                "is not large enough to contain the dilation of the window",
                sQuote("win")))
-  parentlist <- rpoispp(kappa, lmax=lmax, win=dilated, nsim=nsim)
-  if(nsim == 1) parentlist <- list(parentlist)
+
+  parentlist <- rpoispp(kappa, lmax=kappamax, win=dilated,
+                        nsim=nsim, drop=FALSE)
 
   resultlist <- vector(mode="list", length=nsim)
-  for(isim in 1:nsim) {
+  for(isim in seq_len(nsim)) {
     parents <- parentlist[[isim]]
     result <- NULL
     ## generate clusters
@@ -801,10 +807,10 @@ rstrat <- function(win=square(1), nx, ny=nx, k=1, nsim=1, drop=TRUE) {
   stopifnot(k >= 1)
   if(!missing(nsim)) {
     check.1.integer(nsim)
-    stopifnot(nsim >= 1)
+    stopifnot(nsim >= 0)
   }
   result <- vector(mode="list", length=nsim)
-  for(isim in 1:nsim) {
+  for(isim in seq_len(nsim)) {
     xy <- stratrand(win, nx, ny, k)
     Xbox <- ppp(xy$x, xy$y, win$xrange, win$yrange, check=FALSE)
     result[[isim]] <- Xbox[win]
@@ -856,7 +862,8 @@ rcell <- function(win=square(1), nx=NULL, ny=nx, ...,
                   dx=NULL, dy=dx, N=10, nsim=1, drop=TRUE) {
   if(!missing(nsim)) {
     check.1.integer(nsim)
-    stopifnot(nsim >= 1)
+    stopifnot(nsim >= 0)
+    if(nsim == 0) return(simulationresult(list()))
   }
   win <- as.owin(win)
   xr <- win$xrange
@@ -873,7 +880,7 @@ rcell <- function(win=square(1), nx=NULL, ny=nx, ...,
   dy <- g$dy
   ## generate pattern(s)
   result <- vector(mode="list", length=nsim)
-  for(isim in 1:nsim) {
+  for(isim in seq_len(nsim)) {
     x <- numeric(0)
     y <- numeric(0)
     for(ix in seq_len(nx))
@@ -909,36 +916,80 @@ thinjump <- function(n, p) {
 }
 
 rthin <- function(X, P, ..., nsim=1, drop=TRUE) {
-  if(!(is.ppp(X) || is.lpp(X) || is.pp3(X) || is.ppx(X) || is.psp(X)))
+  if(!(inherits(X, c("ppp", "lpp", "pp3", "ppx", "psp")) ||
+       recognise.spatstat.type(X) == "listxy"))
     stop(paste("X should be a point pattern (class ppp, lpp, pp3 or ppx)",
                "or a line segment pattern (class psp)"),
          call.=FALSE)
-  if(!missing(nsim)) {
-    check.1.integer(nsim)
-    stopifnot(nsim >= 1)
+  rthinEngine(X, P, ..., nsim=nsim, drop=drop)
+}
+
+rthinEngine <- function(X, P, ..., nsim=1, drop=TRUE,
+                        Pmax=1, na.zero=FALSE,
+                        what=c("objects", "fate"),
+                        fatal=TRUE, warn=TRUE) {
+  check.1.integer(nsim)
+  stopifnot(nsim >= 0)
+  ## if what = 'objects', return the thinned pattern
+  ## if what = 'fate', return the logical vector (retained/deleted)
+  what <- match.arg(what)
+  ## recognise list(x,y) input
+  if(israwxy <- (recognise.spatstat.type(X) == "listxy")) {
+    xx <- X$x
+    yy <- X$y
+    nX <- length(xx)
+  } else {
+    nX <- nobjects(X)
   }
-  nX <- nobjects(X)
-  if(nX == 0) {
-    result <- rep(list(X), nsim)
+  ## catch trivial cases
+  if(nX == 0 || nsim == 0) {
+    switch(what,
+           objects = {
+             result <- rep(list(X), nsim)
+           },
+           fate = {
+             result <- rep(list(logical(nX)), nsim)
+           })
     result <- simulationresult(result, nsim, drop)
     return(result)
   }
 
+  if(!missing(Pmax)) {
+    check.1.real(Pmax)
+    stopifnot(Pmax > 0)
+  }
+
   if(is.numeric(P) && length(P) == 1 && spatstat.options("fastthin")) {
-    # special algorithm for constant probability
+    ## special fast algorithm for constant probability
+    if(!missing(Pmax)) P <- P/Pmax
+    if(badprobability(P, TRUE)) stop("P is not a valid probability value")
     result <- vector(mode="list", length=nsim)
-    for(isim in 1:nsim) {
-      retain <- thinjump(nX, P)
-      Y <- X[retain]
-      ## also handle offspring-to-parent map if present
-      if(!is.null(parentid <- attr(X, "parentid")))
-        attr(Y, "parentid") <- parentid[retain]
-      result[[isim]] <- Y
-    }
+    switch(what,
+           fate = {
+             for(isim in seq_len(nsim)) 
+               result[[isim]] <- thinjump(nX, P)
+           },
+           objects = {
+             for(isim in seq_len(nsim)) {
+               retain <- thinjump(nX, P)
+               if(israwxy) {
+                 Y <- list(x=xx[retain], y=yy[retain])
+               } else {
+                 Y <- X[retain]
+                 attr(Y, "parents") <- attr(X, "parents")
+                 ## also handle offspring-to-parent map if present
+                 if(!is.null(parentid <- attr(X, "parentid")))
+                   attr(Y, "parentid") <- parentid[retain]
+               }
+               result[[isim]] <- Y
+             }
+           })
     result <- simulationresult(result, nsim, drop)
     return(result)
   }
 
+  #' general case: compute probabilities first
+  
   if(is.numeric(P)) {
     ## vector of retention probabilities
     pX <- P
@@ -948,11 +999,9 @@ rthin <- function(X, P, ..., nsim=1, drop=TRUE) {
       else 
         stop("Length of vector P does not match number of points of X")
     }
-    if(anyNA(pX))
-      stop("P contains NA's")
   } else if(is.function(P)) {
     ## function - evaluate it at points of X
-    if(!(is.ppp(X) || is.lpp(X)))
+    if(!(is.ppp(X) || is.lpp(X) || israwxy))
       stop(paste("Don't know how to apply a function to an object of class",
                  commasep(sQuote(class(X)))),
            call.=FALSE)
@@ -961,36 +1010,74 @@ rthin <- function(X, P, ..., nsim=1, drop=TRUE) {
       stop("Function P returned a vector of incorrect length")
     if(!is.numeric(pX))
       stop("Function P returned non-numeric values")
-    if(anyNA(pX))
-      stop("Function P returned some NA values")
   } else if(is.im(P)) {
     ## image - look it up
-    if(!(is.ppp(X) || is.lpp(X)))
+    if(!(is.ppp(X) || is.lpp(X) || israwxy))
       stop(paste("Don't know how to apply image values to an object of class",
                  commasep(sQuote(class(X)))),
            call.=FALSE)
     if(!(P$type %in% c("integer", "real")))
       stop("Values of image P should be numeric")
     pX <- P[X, drop=FALSE]
-    if(anyNA(pX))
-      stop("some points of X lie outside the domain of image P")
-  } else
-  stop("Unrecognised format for P")
+  } else stop("Unrecognised format for P")
 
-  if(min(pX) < 0) stop("some probabilities are negative")
-  if(max(pX) > 1) stop("some probabilities are greater than 1")
-
-  result <- vector(mode="list", length=nsim)
-  for(isim in 1:nsim) {
-    retain <- (runif(length(pX)) < pX)
-    Y <- X[retain]
-    ## also handle offspring-to-parent map if present
-    if(!is.null(parentid <- attr(X, "parentid")))
-      attr(Y, "parentid") <- parentid[retain]
-    result[[isim]] <- Y
+  if(!all(is.finite(pX))) {
+    if(na.zero) {
+      pX[is.na(pX)] <- 0
+    } else if(anyNA(pX)) {
+      if(is.function(P)) stop("Function P returned some NA values")
+      if(is.im(P)) stop("Some points lie outside the domain of image P")
+      stop("P contains NA's")
+    } else {
+      if(is.function(P)) stop("Function P returned some NaN or Inf values")
+      if(is.im(P)) stop("Image P contains NaN or Inf values")
+      stop("P contains NaN or Inf values")
+    }
   }
+
+  if(fatal || warn) {
+    ## check for bad values of probability
+    ra <- range(pX)/Pmax
+    if(ra[1] < 0) {
+      gripe <- paste("some probabilities are negative",
+                     paren(paste("minimum", ra[1])))
+      if(fatal) stop(gripe)
+      if(warn) stop(gripe)
+    }
+    if(ra[2] > 1) {
+      gripe <- paste("some probabilities are greater than 1",
+                     paren(paste("maximum", ra[2])))
+      if(fatal) stop(gripe)
+      if(warn) warning(gripe)
+    }
+  }
+
+  #' now simulate
+  
+  result <- vector(mode="list", length=nsim)
+  switch(what,
+         fate = {
+           for(isim in seq_len(nsim)) 
+             result[[isim]] <- ((Pmax * runif(length(pX))) < pX)
+         },
+         objects = {
+           for(isim in seq_len(nsim)) {
+             retain <- ((Pmax * runif(length(pX))) < pX)
+             if(israwxy) {
+               Y <- list(x=xx[retain], y=yy[retain])
+             } else {
+               Y <- X[retain]
+               attr(Y, "parents") <- attr(X, "parents")
+               ## also handle offspring-to-parent map if present
+               if(!is.null(parentid <- attr(X, "parentid")))
+                 attr(Y, "parentid") <- parentid[retain]
+             }
+             result[[isim]] <- Y
+             }
+         })
   result <- simulationresult(result, nsim, drop)
   return(result)
 }
 
 
+
